import sys

from PyQt5 import QtCore
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

from .ParserThread import ParserThread
from .ListView import ListView
from ..config import DefinitionLoader
from ..arxiv import filter

class ArxivFilter(QMainWindow):

    def __init__(self, config=None):
        super().__init__()

        self._config = config

        configLoader = DefinitionLoader()
        if self._config is not None:
            configOK = configLoader.loadFromFile(self._config)
        else:
            configOK = configLoader.loadDefault()

        if not configOK:
            error = QErrorMessage()
            if configLoader.error is not None:
                error.showMessage(configLoader.error)
            else:
                error.showMessage("Error while loading the config!")
            error.exec_()

            sys.exit(-1)

        self._filter = filter(configLoader.definition)

        self._loading = False

        self.initUI()

        self.setAcceptDrops(True)

    def initUI(self):

        self.setWindowTitle('Arxiv Daily E-Mail Filter')

        self.lo = QVBoxLayout()

        labelFiltered = QLabel("Filtered Entries", self)
        labelFiltered.setStyleSheet("font-weight: bold; font-size: 12pt");
        labelOther = QLabel("Other Entries", self)
        labelOther.setStyleSheet("font-weight: bold; font-size: 12pt");

        # Construct Filtered Table
        self.tableFiltered = ListView(filtered = True)

        # Construct Other Table
        self.tableOther = ListView(filtered = False)

        # Construct the Usage Hint
        self.hint = QLabel("Drag-and-drop or paste (Ctrl+V) the text of the daily Arxiv E-Mail")
        self.hint.setStyleSheet("font-weight: bold; font-size: 14pt; color: #3e3e3e; background-color: white")
        self.hint.setWordWrap(True)
        self.hint.setAlignment(QtCore.Qt.AlignCenter)
        self._hintShown = True

        self.lo.addWidget(labelFiltered)
        self.lo.addWidget(self.hint)
        self.lo.addWidget(labelOther)
        self.lo.addWidget(self.tableOther)

        self.lo.setStretch(1, 2)
        self.lo.setStretch(3, 1)

        self.lo.insertSpacing(2, 20)

        widget = QWidget()
        widget.setLayout(self.lo)
        self.setCentralWidget(widget)

        # Enable pasting of E-Mail text
        self.shortcutPaste = QShortcut(QKeySequence('Ctrl+V'), self)
        self.shortcutPaste.activated.connect(self.pasteEvent)


    # Drag and Drop handling of E-Mail text
    def dragEnterEvent(self, e):
        if self._loading:
            e.ignore()
        elif e.mimeData().hasText():
            e.accept()
        else:
            e.ignore()

    def dropEvent(self, e):
        text = e.mimeData().text()
        self._setArxivText(text)


    def pasteEvent(self):
        text = QGuiApplication.clipboard().mimeData().text()
        self._setArxivText(text)


    def _setArxivText(self, text):

        self.statusBar().showMessage("Loading ... ")

        self._parser = ParserThread(text, self)
        self._parser.finished.connect(self.parsedCallback)
        self._parser.start()
        self._loading = True


    def _hideHint(self):
        if self._hintShown:
            self.lo.replaceWidget(self.hint, self.tableFiltered)
            self._hintShown = False

    def parsedCallback(self):

        self._hideHint()

        # Clear the status bar and show potential errors
        self.statusBar().clearMessage()

        if self._parser.error() is not None:
            self.statusBar().showMessage(self._parser.error(), 5000)

        filtered = []
        other = []

        for entry in self._parser.entries():
            scored = self._filter.score(entry)

            if scored.score > 0:
                filtered.append(scored)
            else:
                other.append(scored)

        self.tableFiltered.setEntries(filtered)
        self.tableOther.setEntries(other)

        self._loading = False



import yaml
import os

from .Definition import Definition

class DefinitionLoader():

    DEFAULT_LOCATIONS = [
      './',
      '~/.config/',
      '~/',
      '/etc/',
    ]

    FILE_NAME = 'arxiv_filter.yaml'

    def __init__(self):

        self.defintion = Definition({})
        self.error = None

    def loadDefault(self):
        """
        Scans all default config file locations for an existing config file.
        """
        for path in self.DEFAULT_LOCATIONS:
            file = os.path.expanduser(path + self.FILE_NAME)
            if os.path.isfile(file):
                return self.loadFromFile(file)

        self.error = "No default config file (%s) found in [%s]!"%(
            self.FILE_NAME, ', '.join(self.DEFAULT_LOCATIONS))
        return False


    def loadFromFile(self, filename):
        """
        Load the filter definition from given filename

        Arguments:
            filename: The file path to load.

        Returns:
            True if definition succesfully loaded, False otherwise.
        """
        try:
            with open(filename, 'r') as f:
                self.definition = Definition(yaml.safe_load(f))
                return True
        except Exception as e:
            self.error = str(e)
            return False



from .scored_entry import scored_entry
from . import util

class filter():

    def __init__(self, definition) -> None:

        self._definition = definition

    def score(self, entry) -> scored_entry:

        scored = scored_entry(entry)

        for category in self._definition.categories:

            if category.lower() == 'author':
                score, matches = self._scoreList(category, entry.authors)
                if len(matches) > 0:
                    scored.hits['people'] = True
                    scored.score += score
                    scored.matched_authors = matches

            elif category.lower() == 'keyword':
                score, matches = self._scoreString(category, entry.title)
                if len(matches) > 0:
                    scored.hits['title'] = True
                    scored.score += score
                    scored.matched_title = matches

                score, matches = self._scoreString(category, entry.abstract)
                if len(matches) > 0:
                    scored.hits['abstract'] = True
                    scored.score += score
                    scored.matched_abstract = matches

            elif category.lower() == 'category':
                score, matches = self._scoreList(category, entry.categories)
                if len(matches) > 0:
                    scored.hits['category'] = True
                    scored.score += score
                    scored.matched_categories = matches

            elif category.lower() == 'collaboration':
                score, matches = self._scoreString(category, entry.collaboration)
                if len(matches) > 0:
                    scored.hits['group'] = True
                    scored.score += score

        return scored

    def _scoreList(self, category, values) -> tuple[int, list[str]]:
        """
        Score a list of strings according to the given category of the definition.

        Arguments:
            category: The definition category
            values: The list of string to score

        Returns:
            score: The total accumulated score over all entries in values
            matches: List of entries in values which where matched!
        """
        score = 0
        matches = []

        for entry in values:
            clean, clean_mapping = util.saniztize(entry, return_idx=True)
            matched = False

            for item in self._definition.getCategory(category):
                locations = item.findInString(clean, entry, clean_mapping)

                if len(locations) > 0:
                    score += item.value
                    matched = True

            if matched:
                matches.append(entry)

        return score, matches


    def _scoreString(self, category, string) -> tuple[int, list[str]]:
        """
        Score single text (string) according to the given category of the definition.

        Arguments:
            category: The definition category
            string: The text string to score

        Returns:
            score: The total score of the given text
            matches: List of all substring of the text string which were matched
        """
        clean, clean_mapping = util.saniztize(string, return_idx=True)

        score = 0
        matches = set()

        for item in self._definition.getCategory(category):
            locations = item.findInString(clean, string, clean_mapping)

            if len(locations) > 0:
                score += item.value

                for location in locations:
                    start = clean_mapping[location]
                    end = clean_mapping[location+len(item.keyClean)-1] + 1
                    matches.add(string[start:end])

        return score, list(matches)



import time

from PyQt5.QtCore import QThread

from ..arxiv import parser

class ParserThread(QThread):

    def __init__(self, str, parent=None):
        super().__init__(parent)

        self._str = str
        self._parser = parser()

    def entries(self):
        return self._parser.entries()

    def error(self):
        if self._parser.error is not None:
            return self._parser.error
        elif self._parser.warning is not None:
            return self._parser.warning
        else:
            return None

    def run(self):

        self._parser.fromText(self._str)



def saniztize(input_string, return_idx=False):
    """
    Sanitize a text string.

    Two main functions:
    - Remove special (non alphabet characters)
    - Replace characters with umlaut, accents etc. with their non-accentuated version.

    Optionally also returns a list of indexes for the sanitized characters into the original string.

    Example:
        INPUT: input_string = "abc/-dé"
        OUTPUT: clean = "abcde"
                index = [0, 1, 2, 5, 6]

    Arguments:
        input_string: The string to be sanitized
        return_idx: If True, the function also returns a list with an entry for
            character in the sanitized string, whith the index of the sanitized
            character in the original string.

    Returns:
        clean: The cleaned (sanitized) output string
        index: If return_idx is True, the index mapping into the original string.
    """

    # Characters to be removed fully.
    REMOVE = [
        '-', '.', ',', '_', ':', ';',
        '[', ']', '(', ')', '{', '}',
        '^', '\\', '/', '\'', '`', '"', '´',
        '&', '$'
    ]
    # Characters to be replaced
    # Only single string replacements are allowed, otherwise index calculation fails!
    REPLACE = {
        'ä': 'a', 'ö': 'o', 'ü': 'u',
        'é': 'e', 'è': 'e', 'à': 'a', 'â': 'a',
    }

    # Replace characters
    clean = input_string.translate(REPLACE)

    # Remove characters
    index = []
    output = []

    # Create the list of characters to keep
    for cc, char in enumerate(clean):
        if char not in REMOVE:
            output.append(char)
            index.append(cc)

    # Join the kept characters back together!
    clean = ''.join(output).lower()

    if not return_idx:
        return clean
    else:
        return clean, index



import email
import locale

from .entry import entry

class parser:

    FOOTER_DIVIDER = '%%%---%%%---%%%---%%%---%%%---%%%---%%%---%%%---%%%---%%%---%%%---%%%---%%%---'
    ENTRY_DIVIDER = '------------------------------------------------------------------------------\n'


    def __init__(self):
        self._entries = []

        self.error = None
        self.warning = None


    def fromEmailFile(self, filename):
        with open(filename, 'r') as f:
            msg = email.message_from_file(f)

        for part in msg.walk():
            # Select the first text/plain entry
            # There should only be one!
            if part.get_content_type() == 'text/plain':
                return self.fromText(part.as_string())
                break;

    def fromEmail(self, str):
        # TODO implement
        pass


    def fromText(self, str):

        # Set the locale to 'C' in order to correctly parse the (english)
        # dates in the E-mail (within the `entry` class)
        loc = locale.getlocale()
        locale.setlocale(locale.LC_ALL, 'C')

        ## Unify the line-break character
        # Drap-n-Drop and Copy-Paste might not show the same line-break even on linux!
        str = str.replace("\r\n", "\n")

        ## Remove the E-Mail Footer
        if self.FOOTER_DIVIDER not in str:
            self.warning = "Warning: This might not be an arXiv E-Mail (missing footer)."
        else:
            str = str.split(self.FOOTER_DIVIDER)[0]

        splits = str.split(self.ENTRY_DIVIDER)
        if len(splits) < 2:
            self.warning = "Warning: This might not be an arXiv E-Mail (only one entry)."

        for split in splits:
            if not split.startswith('\\\\'):
                continue

            # TODO better error handling on parse fail
            self._entries.append(entry(split))

        locale.setlocale(locale.LC_ALL, loc)

    def entries(self):
        return self._entries;



from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

class EntryLabel(QLabel):

    clicked = pyqtSignal()
    doubleClicked = pyqtSignal()

    def __init__(self, text, parent=None):
        super().__init__(text, parent)

        self.setWordWrap(True)
        self.setStyleSheet("font-weight: bold;");

    def mouseReleaseEvent(self, event):
        self.clicked.emit()

    def mouseDoubleClickEvent(self, event):
        self.doubleClicked.emit()

    def enterEvent(self, event):
        self.setStyleSheet("font-weight: bold; text-decoration: underline;")
        self.setCursor(Qt.PointingHandCursor)

    def leaveEvent(self, event):
        self.setStyleSheet("font-weight: bold; text-decoration: none;")
        self.unsetCursor()


class FoldLabel(QLabel):

    clicked = pyqtSignal()

    def __init__(self, parent=None):
        super().__init__(">", parent)

        self._closed = True

        self.setStyleSheet("font-size: 15pt; font-weight: bold;");

    def toggle(self):
        if self._closed:
            self.setText("v")
        else:
            self.setText(">")

        self._closed = not self._closed

    def mouseReleaseEvent(self, event):
        self.clicked.emit()

    def enterEvent(self, event):
        self.setCursor(Qt.PointingHandCursor)

    def leaveEvent(self, event):
        self.unsetCursor()



class ListEntry(QFrame):

    def __init__(self, entry, filtered=True, parent=None):
        super().__init__(parent)

        self._entry = entry
        self._filtered = filtered

        self.initUI()

    def initUI(self):

        self.setContentsMargins(0, 0, 0, 0)
        self.setFrameStyle(QFrame.Panel | QFrame.Raised);
        self.setLineWidth(1);

        lo = QVBoxLayout(self)

        self._header = QFrame()
        self._content = QWidget()

        # Color the header if filtered
        if self._filtered:
            shade = int(self._entry.score/100 * 255)
            if shade > 255: shade = 255
            
            if self._entry.hits['people']:
                # If an author is matched, color scale: White -> Green
                colors = (255 - shade, 255 - shade/2, 255 - shade/2)
            else:
                # Normal color scale: White -> Blue
                colors = (255 - shade, 255 - shade, 255 - shade/6)

            # Adjust the text color based on the background color
            # To guarantee good contrast
            # Based on https://stackoverflow.com/a/41491220 but with adjusted threshold.
            if ((colors[0] * 0.299) + (colors[1] * 0.587) + (colors[2] * 0.114)) <= 175:
                text_color = '#fff'
            else:
                text_color = '#333'

            self._header.setStyleSheet(f"color: {text_color};background-color: rgb({colors[0]}, {colors[1]}, {colors[2]});");

        self._content.setVisible(False)

        lo.addWidget(self._header)
        lo.addWidget(self._content)

        # Create the header layout
        header_lo = QHBoxLayout()

        self._fold = FoldLabel()
        header_lo.addWidget(self._fold)
        header_lo.setStretch(0, 0)

        title_rich = self._entry.entry.title
        for match in self._entry.matched_title:
            title_rich = title_rich.replace(match, f'<u>{match}</u>')
        title = EntryLabel(title_rich)
        header_lo.addWidget(title)
        header_lo.setStretch(1, 1)

        title.clicked.connect(self.openLink)
        self._fold.clicked.connect(self.toggleContent)

        if self._filtered:
            hits = [key[0].upper() for key, value in self._entry.hits.items() if value]
            reason = QLabel("Score: %i\n\n%s"%(self._entry.score, ''.join(hits)))
            header_lo.addWidget(reason)
            header_lo.setStretch(2, 0)

        self._header.setLayout(header_lo)

        # Create the content
        content_lo = QHBoxLayout()

        ## Authors and categories
        meta_widget = QWidget()
        meta_lo = QVBoxLayout()

        collaboration = QLabel(self._entry.entry.collaboration)
        authors = QLabel('\n\n'.join(self._entry.matched_authors))
        categories = QLabel('\n\n'.join(self._entry.matched_categories))

        collaboration.setStyleSheet("font-weight: bold; font-style: italic")
        authors.setStyleSheet("font-weight: bold")
        categories.setStyleSheet("font-style: italic")

        meta_lo.addWidget(collaboration)
        meta_lo.addWidget(authors)
        meta_lo.addWidget(categories)

        meta_lo.insertSpacing(1, 20)
        meta_lo.insertSpacing(3, 20)

        meta_widget.setLayout(meta_lo)

        ## Abstract
        abstract_rich = self._entry.entry.abstract
        for match in self._entry.matched_abstract:
            abstract_rich = abstract_rich.replace(match, f'<b>{match}</b>')

        abstract = QLabel(abstract_rich)
        abstract.setWordWrap(True)
        abstract.setStyleSheet("font-size: 11pt; line-height: 160%; text-align: justify;")

        content_lo.addWidget(meta_widget)
        content_lo.addWidget(abstract)

        content_lo.setStretch(0, 1)
        content_lo.setStretch(1, 4)

        self._content.setLayout(content_lo)

        self.setLayout(lo)

    def toggleContent(self):
        self._fold.toggle()
        self._content.setVisible(not self._content.isVisible())

    def openLink(self):
        QDesktopServices.openUrl(QUrl(self._entry.entry.link))



from datetime import datetime
import re

class entry:

    DATEFORMAT = '%d %b %Y %H:%M:%S %Z'

    def __str__(self):
        pass

    def __init__(self, entry_str):

        self.abstract = ''
        self.link = ''
        self.size = ''

        self.key = ''
        self.date = None

        self.title = ''

        self.authors = []
        self.collaboration = ''

        self.categories = []
        self.doi = ''
        self.new = True

        parts = entry_str.split('\\\\')

        for part in parts:
            if part.startswith('\narXiv:'):
                self.parseHeader(part)
            elif part.startswith(' ( '):
                self.parseFooter(part)
            else:
                self.parseAbstract(part)


    def parseHeader(self, data):
        # Continuing lines either start with two spaces '  ' or a signal space ' ' after the new-line.
        data = data.replace('\n  ', ' ')
        data = data.replace('\n ', ' ')

        headers = data.split('\n')

        for header in headers:
            header = header.strip()
            if header == '':
                continue

            if header.startswith('replaced with'):
                self.parseDateStr(header)
                self.new = False
                continue

            if header.startswith('Date:'):
                self.parseDateStr(header)
                continue

            try:
                key, value = header.split(':', 1)
                key = key.strip().lower()
                value = value.strip()
            except Exception as e:
                print(header)
                print('Warning: ' + str(e))
                continue

            if key == 'title':
                self.title = self._sanitize(value)
            elif key == 'authors':
                self.parseAuthors(value)
            elif key == 'categories':
                self.categories = value.split(' ')
            elif key == 'doi':
                self.doi = value
            elif key == 'arxiv':
                self.parseKey(value)

            ## TODO support more metadata


    def _sanitize(self, str):
        return str.replace('\\"a', 'ä').replace('\\"u', 'ü').replace('\\"o', 'ö').replace('\\`e', 'è').replace('\\\'e', 'é')


    def parseAbstract(self, data):
        abstract = ' '.join(data.strip().split('\n'))

        # Replace some common latex umlaut notations
        self.abstract = self._sanitize(abstract)


    def parseFooter(self, data):
        data = data.replace('(', '').replace(')', '')

        parts = data.split(',')
        if len(parts) < 2:
            # TOOD raise exception
            return

        self.link = parts[0].strip()
        self.size = parts[1].strip()



    def parseAuthors(self, authorstr):

        authorstr = self._sanitize(authorstr)

        authors = authorstr.strip().split(',')

        # Extract the collaboration at the front
        if ': ' in authors[0]:
            front = authors[0].split(': ', 1)

            self.collaboration = front[0].strip()
            authors[0] = front[1].strip()
        elif 'collaboration' in authors[0].lower():
            self.collaboration = authors[0]
            authors = authors[1:]

        # Further split authors on ' and '
        authors = [part for author in authors for part in author.split(' and ')]

        # This is a very basic author extraction
        # This could very likely be improved in the future for additional
        # edge cases.

        self.authors = [a.strip() for a in authors if len(a.strip()) > 0]

    def parseDateStr(self, datestr):
        try:
            date = re.match(".*, ([^\(]*)\(.*", datestr)
            if date:
                self.date = datetime.strptime(date.groups()[0].strip(), self.DATEFORMAT)
        except Exception as e:
            # TODO better error handling!
            print(e)
            pass

    def parseKey(self, keystr):
        try:
            parts = keystr.strip().split(' ')
            self.key = parts[0]
        except:
            pass



from ..arxiv import util

class DefinitionItem:

    def __init__(self, key, value) -> None:
        self._key = key
        self._exact = False

        # Enable exact matching of a key
        # In this case, the key can not be part of a word!
        if key.startswith('^'):
            self._key = key[1:]
            self._exact = True

        self._clean = util.saniztize(key)
        self._value = value

    @property
    def value(self) -> int:
        return self._value

    @property
    def key(self) -> str:
        return self._key

    @property
    def keyClean(self) -> str:
        return self._clean

    @property
    def isExact(self) -> bool:
        return self._exact

    def _findAll(self, key, string):
        """
        Return list of all indexes of substring `key` in `string`.

        Overlapping matches of `key` are allowed.
        """
        idx = string.find(key)
        while idx != -1:
            yield idx
            idx = string.find(key, idx+1)


    def findInString(self, string_clean, string_original, mapping):
        """
        Find all locations of the `keyClean` in `string_clean`.
        This functions honours the exact matching if requested!

        Exact matching is implemented by checking characters before and after each match
        in the original (non-sanitized) string `string_original`.
        """
        indexes = self._findAll(self.keyClean, string_clean)
        if not self.isExact:
            # If a non-exact match is ok, we can simply return the clean matches
            return list(indexes)
        else:
            exact_indexes = []
            for idx in indexes:
                index_before = mapping[idx] - 1
                index_after = mapping[idx+len(self.keyClean)-1] + 1

                if index_before >=0  and string_original[index_before].isalpha():
                    # Character before the match is alphabet character
                    continue
                if index_after < len(string_original) and string_original[index_after].isalpha():
                    # Character after the match is alphabet character
                    continue

                exact_indexes.append(idx)

            return exact_indexes



class Definition:

    def __init__(self, definiton_data):
        self._data = self._cleanDefinition(definiton_data)

    def _cleanDefinition(self, definiton_data):
        out = {}

        for part, keys in definiton_data.items():
            out[part] = []
            for key, value in keys.items():
                item = DefinitionItem(key, value)
                out[part].append(item)

        return out

    def getCategory(self, category):
        if category in self._data:
            return self._data[category]
        else:
            return {}

    @property
    def categories(self):
        return self._data.keys()



from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

from .ListEntry import ListEntry

class ListView(QScrollArea):

        def __init__(self, filtered=True, parent=None):
            super().__init__(parent)

            self._filtered = filtered
            self._widgets = []

            self.initUI()

        def initUI(self):

            self.setBackgroundRole(QPalette.Light)
            self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
            self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

            widget = QWidget(self)

            self._layout = QVBoxLayout(self)

            widget.setLayout(self._layout)

            self.setWidget(widget)
            self.setWidgetResizable(True) # This line took me ca. 1 hour....

        def clear(self):
            for widget in self._widgets:
                self._layout.removeWidget(widget)
                widget.setParent(None)

            self._widgets = []


        def setEntries(self, entries):
            self.clear()

            if self._filtered:
                scores = [e.score for e in entries]
                indexes = sorted(range(len(scores)), key=scores.__getitem__)
                indexes.reverse()
            else:
                indexes = range(0, len(entries))

            for idx in indexes:
                entry = entries[idx]
                widget = ListEntry(entry, self._filtered)
                self._layout.addWidget(widget)
                self._widgets.append(widget)



import sys
import argparse

from PyQt5.QtWidgets import *
from PyQt5.QtCore import *

from arxiv_filter import ui

def main():
    app = QApplication(sys.argv)

    argv = [str(aa) for aa in app.arguments()]

    # Parse our command line arguments
    parser = argparse.ArgumentParser(description='''
        arXiv Filter
        Simple utility for filtering arXiv daily E-Mail.
        Just drag and drop the E-Mail text into the window and the magic happens :)''')

    parser.add_argument("-c", "--config", type=str, help="Alternative filter config file.")

    args = parser.parse_args(argv[1:])

    window = ui.ArxivFilter(config=args.config)
    window.show()

    app.exec_()

if __name__ == '__main__':
    main()



from .entry import entry

class scored_entry():

    def __init__(self, entry):

        self.entry = entry

        # The keys are chosen, so that all have a different first letter
        # This is used as a shortcut for visualization
        self.hits = {}
        self.hits['people'] = False
        self.hits['group'] = False
        self.hits['title'] = False
        self.hits['abstract'] = False
        self.hits['category'] = False

        self.matched_title = []
        self.matched_abstract = []
        self.matched_authors = []
        self.matched_categories = []

        self.score = 0
